using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using EnvDTE80;
using EnvDTE;
using System.Diagnostics;
using System.Reflection;
using System.Threading;

namespace Com.Benday.VisualStudio.Addins.SolutionSearch
{
    public partial class DocumentOrMethodSearch : Form
    {
        private bool m_isSearching = false;
        private object m_searchSyncRoot = new object();
        private DTE2 m_application = null;
        private List<ProjectItemWrapper> m_allWrappers = null;
        // private List<ProjectItemWrapper> m_matches = null;
        private SortedList<string, ProjectItemWrapper> m_matches = null;

        public DocumentOrMethodSearch(DTE2 dte)
        {
            m_application = dte;
            InitializeComponent();
            AppDomain.CurrentDomain.UnhandledException += new UnhandledExceptionEventHandler(CurrentDomain_UnhandledException);
        }

        void CurrentDomain_UnhandledException(object sender, UnhandledExceptionEventArgs e)
        {
            if (e == null)
            {
                MessageBox.Show("Unhandled exception event arg was null.");
            }
            else if (e.ExceptionObject == null)
            {
                MessageBox.Show("Exception object was null.");
            }
            else
            {
                MessageBox.Show(e.ExceptionObject.ToString());
            }
        }

        private string SafeGetType(object item)
        {
            if (item == null)
            {
                return "(null)";
            }
            else
            {
                return item.GetType().Name;
            }
        }

        public void PopulateFileList()
        {
            if (m_allWrappers == null)
            {
                PopulateFiles();
            }
            else
            {
                System.Threading.Thread tempThread = new System.Threading.Thread(new ThreadStart(PopulateFiles));
                tempThread.Start();
            }
        }

        private void PopulateFiles()
        {
            try
            {
                List<ProjectItemWrapper> allWrappers = new List<ProjectItemWrapper>();
                
                PopulateProjects(m_application.Solution.Projects, 0, allWrappers);

                m_allWrappers = allWrappers;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }
        }

        private void PopulateProjects(Projects projects, int nestLevel, List<ProjectItemWrapper> allWrappers)
        {
            if (projects == null)
            {
                return;
            }

            foreach (Project project in projects)
            {
                PopulateProjectItems(project.ProjectItems, nestLevel + 1, allWrappers);
            }
        }

        private void PopulateProjectItems(ProjectItems projectItems, int nestLevel, List<ProjectItemWrapper> allWrappers)
        {
            if (projectItems == null)
            {
                return;
            }

            StringBuilder sb = new StringBuilder();

            foreach (ProjectItem item in projectItems)
            {
                allWrappers.Add(new ProjectItemWrapper(item));

                if (item.ProjectItems != null)
                {
                    PopulateProjectItems(item.ProjectItems, nestLevel + 1, allWrappers);
                }

                if (item.FileCodeModel != null)
                {
                    PopulateCodeElements(item.FileCodeModel.CodeElements, nestLevel + 1, item, allWrappers);
                }
            }
        }

        private void PopulateCodeElements(CodeElements elements, int nestLevel, ProjectItem item, 
            List<ProjectItemWrapper> allWrappers)
        {
            if (elements == null)
            {
                return;
            }

            StringBuilder sb = null;

            ProjectItemWrapper temp;

            foreach (CodeElement2 element in elements)
            {
                sb = new StringBuilder();

                if (element.Kind == vsCMElement.vsCMElementClass)
                {
                    temp = new ProjectItemWrapper(element, item);
                    allWrappers.Add(temp);

                    // find the project item that has this class in it and remove it from the list --> redundant
                    RemoveProjectItem(temp, allWrappers);
                    PopulateCodeElements(((CodeClass)element).Members, nestLevel + 1, item, allWrappers);
                    PopulateCodeElements(((CodeClass)element).Children, nestLevel + 1, item, allWrappers);                    
                }
                else if (element.Kind == vsCMElement.vsCMElementInterface)
                {
                    allWrappers.Add(new ProjectItemWrapper(element, item));
                }
                else if (element.Kind == vsCMElement.vsCMElementProperty)
                {
                    try
                    {
                        allWrappers.Add(new ProjectItemWrapper(element, item));
                    }
                    catch (Exception ex)
                    {
                        Debug.WriteLine(ex.ToString());
                    }
                }
                else if (element.Kind == vsCMElement.vsCMElementFunction)
                {
                    try
                    {
                        allWrappers.Add(new ProjectItemWrapper(element, item));
                    }
                    catch(Exception ex)
                    {
                        Debug.WriteLine(ex.ToString());
                    }
                }
                else if (element.Kind == vsCMElement.vsCMElementNamespace)
                {
                    PopulateCodeElements(((CodeNamespace)element).Members, nestLevel + 1, item, allWrappers);
                }
                else if (element.Kind == vsCMElement.vsCMElementVariable)
                {
                 
                }
                else
                {
                    sb.Append('-', nestLevel);
                    sb.Append(element.Kind.ToString());
                    Debug.WriteLine(sb.ToString());
                }

            }
        }

        private void RemoveProjectItem(ProjectItemWrapper classItem, List<ProjectItemWrapper> allWrappers)
        {
            if (classItem == null || allWrappers == null)
            {
                return;
            }

            ProjectItemWrapper removeThis = null;

            foreach (ProjectItemWrapper wrapper in allWrappers)
            {
                if (wrapper.ItemType == "PROJECT_ITEM" && 
                    wrapper.SafeGetPathFromProjectItem() == classItem.SafeGetPathFromProjectItem())
                {
                    removeThis = wrapper;
                    break;
                }
            }

            if (removeThis != null) allWrappers.Remove(removeThis);
        }

        private void m_textSearch_TextChanged(object sender, EventArgs e)
        {
            try
            {
                StartSearch(m_textSearch.Text);
            }
            catch (Exception ex)
            {
                this.BringToFront();
                MessageBox.Show(ex.ToString());
            }
        }

        private string m_currentSearch = String.Empty;
        private string m_requestedSearch = String.Empty;

        private void StartSearch(string searchString)
        {
            m_requestedSearch = m_textSearch.Text.Trim().ToLower();
        }

        private void m_timer_Tick(object sender, EventArgs e)
        {
            try
            {
                if (m_requestedSearch != null &&
                    m_currentSearch != m_requestedSearch &&
                    m_background.IsBusy == false &&
                    m_isSearching == false)
                {
                    m_background.RunWorkerAsync(m_requestedSearch.Trim());
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }
        }

        private void m_background_DoWork(object sender, DoWorkEventArgs e)
        {
            m_isSearching = true;

            try
            {
                string search = e.Argument as string;
                m_currentSearch = search;

                // m_matches = new List<ProjectItemWrapper>();
                SortedList<string, ProjectItemWrapper> matches = new SortedList<string, ProjectItemWrapper>();

                string key;
                
                if (search != null && m_allWrappers != null)
                {
                    search = search.ToLower();
                    
                    foreach (ProjectItemWrapper wrapper in m_allWrappers)
                    {
                        if (m_background.CancellationPending == true)
                        {
                            break;
                        }
                        else
                        {
                            if (wrapper.Name != null && wrapper.Name.Contains(search) == true)
                            {
                                key = String.Format("{0}:{1}", wrapper.Name, wrapper.ItemType);

                                if (matches.ContainsKey(key) == true)
                                {
                                    key = key + " " + Guid.NewGuid().ToString();
                                }

                                matches.Add(key, wrapper);
                            }
                        }
                    }
                    
                }

                m_matches = matches;
                
            }
            catch (Exception ex)
            {
                this.BringToFront();
                MessageBox.Show(ex.ToString());
            }
            finally
            {
                m_isSearching = false;
            }
        }

        private void m_background_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            try
            {
                // m_list.DataSource = m_matches.Values;                
                m_bindingSource.DataSource = m_matches.Values;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }            
        }

        private void DocumentOrMethodSearch_Activated(object sender, EventArgs e)
        {
            
        }

        private void DocumentOrMethodSearch_Load(object sender, EventArgs e)
        {
            try
            {
                m_textSearch.Text = String.Empty;
                m_textSearch.Focus();
                this.Visible = true;
            }
            catch (Exception ex)
            {
                this.BringToFront();
                MessageBox.Show(ex.ToString());
            }
        }

        private void m_btnRefresh_Click(object sender, EventArgs e)
        {
            try
            {
                PopulateFileList();
            }
            catch (Exception ex)
            {
                this.BringToFront();
                MessageBox.Show(ex.ToString());
            }
        }

        private void m_textSearch_KeyDown(object sender, KeyEventArgs e)
        {
            try
            {
                if (e.KeyCode == Keys.Down || e.KeyCode == Keys.Up)
                {
                    if (m_list.Items.Count > 0)
                    {
                        m_list.SelectedIndex = 0;
                        m_list.Focus();
                    }
                }
                else if (e.KeyCode == Keys.Enter)
                {
                    ProjectItemWrapper wrapper = m_list.SelectedItem as ProjectItemWrapper;

                    if (wrapper != null)
                    {
                        NavigateToWrapper(wrapper);
                    }
                }
            }
            catch (Exception ex)
            {
                this.BringToFront();
                MessageBox.Show(ex.ToString());
            }
        }

        private void m_list_KeyDown(object sender, KeyEventArgs e)
        {
            try
            {
                if (e.KeyCode == Keys.Enter)
                {
                    ProjectItemWrapper wrapper = m_list.SelectedItem as ProjectItemWrapper;

                    if (wrapper != null)
                    {
                        NavigateToWrapper(wrapper);
                    }
                }
                else if (e.KeyCode != Keys.Down && e.KeyCode != Keys.Up &&
                    e.KeyCode != Keys.ShiftKey)
                {
                    if (e.KeyCode == Keys.Back)
                    {
                        string text = m_textSearch.Text;

                        if (text.Length > 0)
                        {
                            m_textSearch.Text = text.Remove(text.Length - 1);
                            m_textSearch.SelectionStart = m_textSearch.Text.Length;
                        }
                    }
                    else if (e.Shift == true)
                    {
                        m_textSearch.AppendText(e.KeyCode.ToString());
                    }
                    else
                    {
                        m_textSearch.AppendText(e.KeyCode.ToString().ToLower());
                    }

                    m_textSearch.Focus();
                }
            }
            catch (Exception ex)
            {
                this.BringToFront();
                MessageBox.Show(ex.ToString());
            }
        }

        private void NavigateToWrapper(ProjectItemWrapper wrapper)
        {
            try
            {
                Document targetDocument = null;

                if (wrapper != null)
                {
                    if (targetDocument == null)
                    {
                        m_application.ExecuteCommand("File.OpenFile", wrapper.Item.get_FileNames(0));

                        foreach (Document doc in m_application.Documents)
                        {
                            if (doc.FullName == wrapper.Item.get_FileNames(0))
                            {
                                targetDocument = doc;
                                doc.Activate();
                                break;
                            }
                        }
                    }
                }

                if (targetDocument != null)
                {
                    if (wrapper.ItemType == "CLASS")
                    {
                        targetDocument.Activate();
                    }
                    else if (wrapper.ItemType == "PROPERTY" || wrapper.ItemType == "FUNCTION")
                    {
                        
                        bool result = wrapper.TextPoint.TryToShow(
                            vsPaneShowHow.vsPaneShowCentered, System.Reflection.Missing.Value);

                        TextSelection selection = (TextSelection)targetDocument.Selection;
                        selection.MoveToPoint(wrapper.TextPoint, false);
                        

                        //MessageBox.Show(result.ToString());

                    }
                }

                this.Hide();
            }
            catch (ApplicationException ex)
            {
                this.BringToFront();
                MessageBox.Show(ex.ToString());
            }
        }

        private void m_list_KeyPress(object sender, KeyPressEventArgs e)
        {

        }

        private void DocumentOrMethodSearch_KeyDown(object sender, KeyEventArgs e)
        {
            
        }

        private void m_list_SelectedIndexChanged(object sender, EventArgs e)
        {

        }

        private void m_btnCancel_Click(object sender, EventArgs e)
        {
            this.Visible = false;
        }

        public void ClearSearch()
        {
            m_textSearch.Text = String.Empty;
            // m_list.DataSource = m_allWrappers;
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            About about = new About();
            about.ShowDialog(this);
        }

        private void DocumentOrMethodSearch_HelpButtonClicked(object sender, CancelEventArgs e)
        {
            About about = new About();
            about.ShowDialog(this);
        }

        private void m_list_DoubleClick(object sender, EventArgs e)
        {
            try
            {
                ProjectItemWrapper wrapper = m_list.SelectedItem as ProjectItemWrapper;

                if (wrapper != null)
                {
                    NavigateToWrapper(wrapper);
                }
            }
            catch (ApplicationException ex)
            {
                MessageBox.Show(ex.ToString());
            }
        }

        private void m_linkAbout_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            About about = new About();
            about.ShowDialog(this);
        }

        private void m_linkRefresh_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            try
            {
                PopulateFileList();
            }
            catch (Exception ex)
            {
                this.BringToFront();
                MessageBox.Show(ex.ToString());
            }
        }
    }
}